/* DCache.H
 * Private/Shared Data Cache for multithreaded processors
 * Parameters:
 *   geometry, associativity,
 *   policy (write-through/back, write-allocate/no-allocate)
 */

#ifndef AINT_DCACHE_H
#define AINT_DCACHE_H

#include <sys/types.h>

extern inline int Log2(ulong);

// Cache Policy bit masks
const CP_WriteThrough  = 0x1;
const CP_WriteAllocate = 0x2;

class DCache {
public:
  int Height;
  int Width;
  int Associativity;

  uint Policy;

  ushort TagShift;
  ushort IndexBits;
  ushort IndexShift;

  ulong *Tags;
  ushort *Valid;
  ushort *Dirty;
  ushort *Owner;
  ulong *TimeStamp;

  ulong Intra_Reuse; // Read Use by inst in same thread
  ulong Inter_Reuse; // inst in another thread
  ulong Intra_Conflicts; // replaced by inst in same thread
  ulong Inter_Conflicts; // replaced by inst in another thread

  ulong Intra_Rewrites;  // Overwritten by inst from same thread
  ulong Inter_Rewrites;  // from another thread
  ulong Intra_WrConfs;  // Write miss from same thread
  ulong Inter_WrConfs;  // from another thread

  ulong Refs;

  ulong Reads;
  ulong ReadMisses;
  ulong Writes;
  ulong WriteMisses;
  ulong WritesToMem;

public:
  DCache(int height, int width, int associativity, uint policy)
  {
    int ArraySize;

    Height=height;
    Width=width;
    Associativity=associativity;

    Policy = policy;

    IndexBits = Log2(Height);
    IndexShift = Log2(Width);
    TagShift = IndexBits+IndexShift;
    
    ArraySize=Height*Associativity;

    Tags = new ulong [ArraySize];
    Valid = new ushort [ArraySize];
    Dirty = new ushort [ArraySize];
    Owner = new ushort [ArraySize];

    TimeStamp = new ulong [ArraySize];

    Reads = 0;
    ReadMisses = 0;
    Writes = 0;
    WriteMisses = 0;
    WritesToMem=0;

    Intra_Reuse = 0;
    Inter_Reuse = 0;
    Intra_Conflicts = 0;
    Inter_Conflicts = 0;

    Intra_Rewrites = 0;
    Inter_Rewrites = 0;
    Intra_WrConfs = 0;
    Inter_WrConfs = 0;

    Refs = 0;
  }

  ~DCache()
  {
    delete Tags;
    delete Valid;
    delete Dirty;
    delete Owner;
  }

  int TagMatch(int index, ulong tag)
  {
    // Return way for which tags match
    int way;
    for (way=0; way<Associativity; way++) {
      if (Tags[index*Associativity+way] == tag) {
	if (Valid[index*Associativity+way]) {
	  return way;
	}
      }
    }
    return -1;
  }

  int Victim(int index)
  {
    // Find unused or least recently used way
    int way, minway;
    ulong mintime = (ulong)(1ul<<63);
    ulong ts;

    for (way=0; way<Associativity; way++) {
      if (!Valid[index*Associativity+way]) return way;
      if ( (ts = TimeStamp[index*Associativity+way]) < mintime) {
	mintime=ts;
	minway=way;
      }
    }
    return minway;
  }

  void Read(ulong addr, int tid)
  {
    ulong tag = addr >> TagShift;
    int index = (addr >> IndexShift) & (Height-1);
    int way;

    Reads++;
    Refs++;

    // Look up the tag in the tag array
    if ( (way = TagMatch(index, tag)) == -1 ) {
      ulong iaw;
      // This is a miss
      ReadMisses++;

      // Find a victim and kill it
      way = Victim(index);

      iaw = index*Associativity+way;

      if (Valid[iaw]) {
        if (Owner[iaw] == tid) {
	  Intra_Conflicts++;
        } else {
	  Inter_Conflicts++;
        }
      }

      // Is this a write-back cache? If yes, write back
      if (!(Policy & CP_WriteThrough)) {
	if (Dirty[iaw]) {
	  WritesToMem++;
	}
      }


      Tags[iaw] = tag;
      Valid[iaw] = 1;
      Dirty[iaw] = 0;
      Owner[iaw] = tid;

      // Update the timestamp on this way
      TimeStamp[index*Associativity+way] = Refs;

    } // if miss
    else {
      if (Owner[index*Associativity + way] == tid) {
        Intra_Reuse++;
      } else {
        Inter_Reuse++;
      }
    } // else - if hit
  }

  void Write(ulong addr, int tid)
  {
    ulong tag = addr >> TagShift;
    int index = (addr >> IndexShift) & (Height-1);
    int way;

    Writes++;
    Refs++;

    if ( (way = TagMatch(index, tag)) == -1) {
      ulong iaw = index*Associativity+way;

      // Write Miss
      WriteMisses++;

      // Is this a WriteAllocate Cache? If yes, get line
      if (Policy & CP_WriteAllocate) {
	way = Victim(index);

	// Is this a write-back cache? If yes, write back
	if (!(Policy & CP_WriteThrough)) {
	  if (Dirty[iaw] && Valid[iaw]) {
	    WritesToMem++;
            if (Owner[iaw] == tid) {
              Intra_WrConfs++;
            } else {
              Inter_WrConfs++;
            }
	  }
	}

	Tags[iaw] = tag;
	Valid[iaw] = 1;
	Dirty[iaw] = 1;
	TimeStamp[iaw] = Refs;
      } else {
	// Write no-allocate. Just write to memory
	WritesToMem++;
      }
    } else {
      // Hit. If write-through, write to mem
      if (Policy & CP_WriteThrough) {
	WritesToMem++;
      } else {
	Dirty[index*Associativity+way] = 1;
      }
      if (Owner[index*Associativity+way] == tid) {
	Intra_Rewrites++;
      } else {
	Inter_Rewrites++;
      }
      Owner[index*Associativity+way] = tid;
      TimeStamp[index*Associativity+way] = Refs;
    }
  }

};


#endif
