/* ResSim.C
 * Simulator for private/shared resources in a multithreaded simulator
 * Resources are:
 *   Data Cache
 *   Branch Predictor
 *   Branch Target Buffer
 */

#include <iostream.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <aint.h>

#include <DCache.H>
#include <BrPred.H>
#include <BTB.H>

// Prototype declaration
void one_cycle(int tid);

const RS_MaxThreads = 32;
int RS_ActiveThreads;
int RS_UsedThreads;

int Br_Shared;
int PredSize;
int bhrsize;
int BTBSize;
int dc_associativity;
int DCflags;
int dcsize;
int data_is_pc;
BTB_TAG_type BTBtagtype;

/* Global Variables */
char ThreadActiveMask[RS_MaxThreads];
aint_addr_t NextPC[RS_MaxThreads];

DCache *DC   [RS_MaxThreads];
BrPred *Pred [RS_MaxThreads];
BTB    *btb  [RS_MaxThreads];

/* Usage string printed out in sim_usage */
char *RS_Usage = "\n\
Multithreaded Resource Usage Simulator\n\
Usage:\n\
  ResSim <aint-options> -- <ResSim options> <Object File>\n\
ResSim option:\n\
  -s    Branch Predictor and BTB Shared\n\
  -p    Branch Predictor and BTB Private (default)\n\
  -P n  Predictor size (n counters)\n\
  -H n  Branch History Register size (bits) \n\
  -B n  BTB size (n lines)\n\
  -c    BTB has cache line tags (default - false)\n\
  -C    BTB stores cache line numbers (default - false) \n\
  -D n  Dcache size\n\
  -w [t|b]  Write Policy (write-Through/write-Back) (default: wr thru)\n\
  -W [a|n]  Allocate policy (write-Allocate/write-No-allocate) (default - allocate)\n\
  -A n  DCache associativity (default 1 - direct-mapped)\n";


int
sim_init(int argc, char **argv, char **envp)
{
  extern int optind;
  int tid;
  char c;

  RS_ActiveThreads = 0;
  RS_UsedThreads = 0;
  for (tid=0; tid<RS_MaxThreads; tid++) {
    ThreadActiveMask[tid] = 0;
  }

  // Default values for parameters
  Br_Shared = 0;
  PredSize = 256;
  BTBSize = 256;
  BTBtagtype = BTB_TAG_PC;
  data_is_pc = 1;
  dcsize = 256;
  dc_associativity = 1;
  DCflags = CP_WriteThrough|CP_WriteAllocate;
  bhrsize = 5;

  // Process the args here
  while ( (c = getopt(argc, argv, "spP:H:B:cCD:w:W:A:")) != -1 ) {
    switch(c) {
    case 's':
      Br_Shared = 1;
      break;
    case 'p':
      Br_Shared = 0;
      break;
    case 'P':
      PredSize = atoi(optarg);
      break;
    case 'H':
      bhrsize = atoi(optarg);
      break;
    case 'B':
      BTBSize = atoi(optarg);
      break;
    case 'c':
      BTBtagtype = BTB_TAG_CacheIndex;
      break;
    case 'C':
      data_is_pc = 0;
      break;
    case 'D':
      dcsize = atoi(optarg);
      break;
    case 'w':
      if (optarg[0]=='t')
	DCflags |= CP_WriteThrough;
      else
	DCflags &= (~CP_WriteThrough);
      break;
    case 'W':
      if (optarg[0] == 'a')
	DCflags |= CP_WriteAllocate;
      else
	DCflags &= (~CP_WriteAllocate);
      break;
    case 'A':
      dc_associativity = atoi(optarg);
      break;
    default:
      break;
    }
  }

  DC[0] = new DCache(dcsize, 32, dc_associativity, DCflags);

  // Not simulating icache. Assume icache is 1k*32bytes -> 10bit index

  if (Br_Shared) {
    Pred[0] = new BrPred(PredSize, bhrsize);
    btb[0] = new BTB(BTBSize, BTBtagtype, data_is_pc, 10);
  } else {
    int i;
    for (i=0; i<RS_MaxThreads; i++) {
      Pred[i] = new BrPred(PredSize, bhrsize);
      btb[i] = new BTB(BTBSize, BTBtagtype, data_is_pc, 10);
    }
  }
 
  cerr << endl
       << "Simulating Predictor, BTB and DCache:" << endl
       << (Br_Shared?"Shared":"Private") << " Branch hardware" << endl
       << "Predictor size: " << PredSize << endl
       << "BTB size: " << BTBSize << endl
       << "DCache size: " << dcsize << " lines * 32 byte ways, "
       <<  dc_associativity << "-way associative" << endl;


  return optind;
}


void
sim_thread_begin(int tid, aint_addr_t start_pc)
{
  // Create a path
  AINT_create_path(tid, start_pc);

  NextPC[tid] = start_pc;

  RS_ActiveThreads++;
  RS_UsedThreads++;
  ThreadActiveMask[tid] = 1;
}


void
sim_usage(int argc, char **argv)
{
  cerr << endl << RS_Usage << endl;
}

void
sim_thread_end(int tid)
{
  RS_ActiveThreads--;
  ThreadActiveMask[tid] = 0;
}

void
sim_thread_block(int tid)
{
  RS_ActiveThreads--;
  ThreadActiveMask[tid] = 0;
}

void
sim_thread_unblock(int tid)
{
  RS_ActiveThreads++;
  ThreadActiveMask[tid] = 1;
}


void
sim_execute()
{
  int tid;
  while(RS_ActiveThreads) {
    // Find the next active thread
    for (tid=0; tid<RS_UsedThreads; tid++) {
      if (ThreadActiveMask[tid]) {
        one_cycle(tid);
      }
    }
  }
}

void
one_cycle(int tid)
{
  thread_ptr pthread=Threads[tid];
  path_ptr ppath = pthread->path_tree;

  inflight_inst_ptr ifi = AINT_fetch(ppath, NextPC[tid]);

  AINT_issue(ifi);

  if (ifi->iflags & E_CONDBR) {
    // Refer to the predictor and BTB
    int Taken = (ifi->nextPC == ifi->picode->target->addr);
    ulong br_target;
    int btbhit;

    if (Br_Shared) {
      Pred[0]->Predict(NextPC[tid]);
      Pred[0]->Update(NextPC[tid], Taken);
      br_target = btb[0]->Lookup(NextPC[tid]);
      if (data_is_pc) btbhit = (br_target == ifi->nextPC);
      else btbhit = 1; // this needs to compare actual cache line #s
      btb[0]->Update(NextPC[tid], ifi->nextPC, btbhit);
    } else {
      Pred[tid]->Predict(NextPC[tid]);
      Pred[tid]->Update(NextPC[tid], Taken);
      br_target = btb[tid]->Lookup(NextPC[tid]);
      if (data_is_pc) btbhit = (br_target == ifi->nextPC);
      else btbhit = 1; // this needs to compare actual cache line #s
      btb[tid]->Update(NextPC[tid], ifi->nextPC, btbhit);
    }

  } else if (ifi->iflags & E_READ) {
    
    // dcache read
    DC[0]->Read(ifi->vaddr, tid);
    
    AINT_do_read(ifi);
  } else if (ifi->iflags & E_WRITE) {

    // dcache write
    DC[0]->Write(ifi->vaddr, tid);

    AINT_do_write(ifi);
  }  

  NextPC[tid] = ifi->nextPC;
  if (NextPC[tid] == 0) ThreadActiveMask[tid] = 0;

  AINT_commit(ifi);
}


void
sim_done()
{
  // Print Stats

  // Branch Predictor and BTB Accuracy:
  if (Br_Shared) {
    printf("BPR> %10ld %10ld %0.6f\n", Pred[0]->Refs, Pred[0]->Mispredictions, (double)Pred[0]->Mispredictions/Pred[0]->Refs);
    printf("BTB> %10ld %10ld %0.6f\n", btb[0]->Refs, btb[0]->Misses, (double)btb[0]->Misses/btb[0]->Refs);
  } else {
    ulong PredRefs=0;
    ulong Mispreds = 0;
    ulong BTBRefs = 0;
    ulong BTBMisses =0;
    int i;
    for (i=0; i<RS_UsedThreads; i++) {
      PredRefs += Pred[i]->Refs;
      Mispreds += Pred[i]->Mispredictions;
      BTBRefs += btb[i]->Refs;
      BTBMisses += btb[i]->Misses; 
    }
    printf("BPR> %10ld %10ld %0.6f\n", PredRefs, Mispreds, (double)Mispreds/PredRefs);
    printf("BTB> %10ld %10ld %0.6f\n", BTBRefs, BTBMisses, (double)BTBMisses/BTBRefs);
  }
  printf("DCR> %10ld %10ld %0.6f\n", DC[0]->Reads, DC[0]->ReadMisses, (double)DC[0]->ReadMisses/DC[0]->Reads);

  printf("DCW> %10ld %10ld %0.6f\n", DC[0]->Writes, DC[0]->WriteMisses, (double)DC[0]->WriteMisses/DC[0]->Writes);

  ulong Total_Conflicts = DC[0]->Intra_Conflicts + DC[0]->Inter_Conflicts;
  ulong Total_WrConfs = DC[0]->Intra_WrConfs + DC[0]->Inter_WrConfs;
  if (Total_Conflicts == 0) Total_Conflicts = 1;
  if (Total_WrConfs == 0) Total_WrConfs = 1;

  ulong Total_Reuse = DC[0]->Intra_Reuse + DC[0]->Inter_Reuse;
  if (Total_Reuse == 0) Total_Reuse = 1;

  ulong Total_Rewrites = DC[0]->Intra_Rewrites + DC[0]->Inter_Rewrites;
  if (Total_Rewrites == 0) Total_Rewrites = 1;

  printf("DRI> %10ld (%0.6f) %10ld (%0.6f)\n",
    DC[0]->Intra_Reuse, (double) DC[0]->Intra_Reuse/Total_Reuse,
    DC[0]->Intra_Conflicts, (double)DC[0]->Intra_Conflicts/Total_Conflicts);

  printf("DRX> %10ld (%0.6f) %10ld (%0.6f)\n",
    DC[0]->Inter_Reuse, (double) DC[0]->Inter_Reuse/Total_Reuse,
    DC[0]->Inter_Conflicts, (double) DC[0]->Inter_Conflicts/Total_Conflicts);

  printf("DWI> %10ld (%0.6f) %10ld (%0.6f)\n",
    DC[0]->Intra_Rewrites, (double)DC[0]->Intra_Rewrites/Total_Rewrites,
    DC[0]->Intra_WrConfs, (double) DC[0]->Intra_WrConfs/Total_WrConfs);

  printf("DWX> %10ld (%0.6f) %10ld (%0.6f)\n",
    DC[0]->Inter_Rewrites, (double) DC[0]->Inter_Rewrites/Total_Rewrites,
    DC[0]->Inter_WrConfs , (double) DC[0]->Inter_WrConfs/Total_WrConfs);

}


inline
int
Log2(ulong Value)
{
  int i;
  for (i=0; Value>>i; i++);
  return i-1;
}
