/* path.h
 * A path is a control stream that forms a node in the path-tree of
 * a thread. The prime path (root of this tree) begins at the first
 * in-flight instruction in the thread. Other paths, if any, are
 * created at diffident branches, when they are fetched (?)
 *   * Vega 10/03/96: It may happen that there are no resources to
     * create a new path at the time a diffident branch is fetched.
     * however, resources may become available at a later time, but
     * before the branch is resolved/killed. Can a new path be created
     * at that time?
 * Each path has a pc to fetch from, or pointers to its taken and not-taken
 * children. Each path has a register map into the physical registers
 * of it's owning thread. Each path is identified by a unique path_id
 * that also completely specifies its position in the path tree.
 *
 */

#ifndef PATH_H
#define PATH_H

#include "protos.h"

// The path-id structures
const pathid_t ROOT_PATH_ID = (pathid_t) 0x1;

inline pathid_t GET_LCHILD(pathid_t pathid) { return (pathid << 1); }
inline pathid_t GET_RCHILD(pathid_t pathid) { return (pathid << 1) + 1; }
inline pathid_t LEFTCHILD(pathid_t pathid) { return pathid*2; }
inline pathid_t RIGHTCHILD(pathid_t pathid) { return pathid*2 + 1; }
inline int IS_ANCESTOR(pathid_t _i, pathid_t _j) {
  pathid_t _k = _j;
  while(_k > _i) _k = _k >> 1;
  return (_k == _i);
}

//FIXME - Change this to hand-code the O(log n) first set-bit algorithm
//FIXME - is this ever needed?
inline int FSB(ulong _n) {int i; ulong _k = _n; for(i=0; _k; i++, _k=_k>>1); return i;}


// Size of the inflight-inst window from this path.
// This must be a power of 2.
class InstWin;

class path {
public:
  path(pathid_t my_id, path_ptr _parent, thread_ptr _pthread, aint_addr_t _pc);
  ~path();

  pathid_t pathid;

  // Is this really necessary?
  int path_num;

  Physical_RegNum RegMap[2*N_Logical_Regs];

  long &pMapReg(Logical_RegNum lreg)
    { return pthread->Reg[RegMap[lreg]].Int64; }

  double &pMapFP(Logical_RegNum lfp)
    { return pthread->Reg[RegMap[FP_REDIRECT_Z(lfp)]].Double; }

  void promote(pathid_t new_pathid);

  class InstWin *instwin;

  thread_ptr pthread;

  aint_addr_t pc;
  icode_ptr picode;

  // the left and right designation is arbitrary
  path_ptr left;
  path_ptr right;
  path_ptr parent;

  void *priority; // generic type
};

void promote_path_tree(path_ptr ppath);

#endif
