/* aint_init.C
 * initialize aint structures, parse args, call back-end initializers
 */


// put in #includes here

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <time.h>
#include <signal.h>

//these includes should be clubbed into one aint.h file
#include "aint_types.h"
#include "aint_consts.h"

// all EXTERN globals are declared here
#define AINT_INIT
#include "globals.h"

#include "aint.h"
#include "context.h"
#include "protos.h"

// prototypes of sim_ functions
#include "sim.h"

// prototypes of functions used in this file
static void save_args(int argc, char **argv);
static void parse_args(int argc, char **argv);
static void usage(int argc, char **argv);
static void init_consts(void);

unsigned long find_proc_addr(const char *Objname, const char *symbol_name);

int Argc;
char **Argv;

void
aint_init(int argc, char **argv, char **envp)
{

  // used to keep track of where we are in the arg list
  int next_arg;

  Start_time = time(NULL);
  Mem_size_start = (ulong) sbrk(0);

  // make GNU getopt behave as expected
  putenv("POSIXLY_CORRECT=1");

  Simname = argv[0];

  // Set all constants to their initial values
  init_consts();

  // paranoia inherited from mint. might need this later
  save_args(argc, argv);

  /* command line arguments come in four categories:
   *  arguments to aint (parse args)
   *  arguments to the back-end simulator (sim_init)
   *  names of the executable objects (aint_init)
   *  arguments to the executable objects (copy_argv)
   *
   * parse_args() parses arguments to aint, and stops when it finds
   * the first job separator ("--")
   */

  parse_args(argc, argv);
 
  /* if, for whatever reason, the simulator doesn't want
   * the job separator to be "--", it should set its value to
   * some other string using the -j argument to aint
   */
  if (!job_separator)
    job_separator = "--";

  /* The next set of arguments are arguments to the backend simulator
   * if any, followed by the object file names and arguments to object
   * programs
   * 
   * The backend must parse all arguments intended for it and maintain
   * a valid value of (extern) optind. If there are no arguments,
   * the backend needs to do nothing.
   */
   
  next_arg = sim_init(argc, argv, envp);

  if (next_arg >= argc) {
    error("missing object file.");
    usage(argc, argv);
    sim_usage(argc, argv);
    exit(1);
  }

  /* For each job specified, start a new process, load the executable,
   * and leave it in a form suitable for beginning execution
   */
  while(next_arg < argc) {
 
    char *Objname = Argv[next_arg]; 

    /* Create a new context for this file, and load and initialize it
     * The context consists of a new process, thread and path.
     */


    // Creating the process automatically creates the main thread,
    // loads the object file, and initializes the address space

    // Similarly, creating a thread automatically creates the main path in the thread

    process_ptr pprocess = new process(Objname);


    /* A major problem faced by the aint execution methodology is that
     * while events (in the discrete event simulation) can be triggered
     * by certain specific classes of operations, such as instructions
     * and function calls, this alone is not a sufficiently general
     * handle on "macro" events, such as synchronization operations
     * The sim_user function provides a generalized control mechanism
     * from the simulated executable into the simulator.
     * Aint traps the sim_user function (the name can be specified on the
     * command line using the -<FIXME> argument), and substitutes the
     * aint-space sim_user (name fixed) function in its place. This function
     * can access structures in aint and the backend (such as global vars
     * and pointer derefs), thus realizing the object-to-simulator feedback.
     */
    pprocess->sim_user_addr = find_proc_addr(Objname, "sim_user");
    if (pprocess->sim_user_addr) {
      /*
       * The entry in the symbol table is from the "globally callable"
       * symbol (i.e., the true entry to sim_user..).
       */
      //if (!Quiet)
      informative("found_callbacks from " << Objname << endl);
    }


    /* This function probes the innards of the object file manually
     * and prints out relevant information about headers, sections
     * and other assorted organs. Retained here for debug purposes
     * only
     */
    if (Verbose) {
      // read_hdrs(Objname);
    }
    
    // Copy argc, argv, envp onto the process stack
    next_arg = pprocess->copy_argv(next_arg, argc, argv, envp);

  } // repeat for the next executable


  // Handle signals during issue
  struct sigaction action, oaction;

  action.sa_handler = (void(*)(int))handle_signals_for_issue;
  action.sa_mask = 0;
  action.sa_flags = 0;
  sigaction(SIGFPE, &action, &oaction);
  AINT_in_issue = 0;

}


static void
usage(int argc, char **argv)
{
  // print out a detailed list of aint command line options here
  cerr << "\nAINT Usage: " << argv[0] << " <AINT options> ...\n"
       << "    AINT options:\n"
       << "    -h        Help (this message)\n"
       << "    -j js     Use <js> as job separator (default: -- )\n"
       << "    -v        Verbose (lots of diagnostics)\n"
       << "    -x addr   call ObjHookProc when fetching from addr - debug aid" << endl;
}


static void
init_consts(void)
{
  ProcessCount = 0;
  ThreadCount = 0;
  PathCount = 0;
  FetchCount = 0;

  Stack_size = STACK_SIZE;
  Heap_size = HEAP_SIZE;
  Shmem_size = SHMEM_SIZE;

  Verbose = 0;
  Print_Warnings = 0;
}


static void
save_args(int argc, char **argv)
{
  Argc = argc;
  Argv = new char * [Argc + 1];

  // argv[argc] is null
  for (int i=0; i<=argc; i++) {
    Argv[i] = strdup(argv[i]);
  }

}



static void
parse_args(int argc, char **argv)
{
  int c, errflag;
  extern char *optarg;
  extern int optind;


  /* ObjHook is used in next_event() to insert a call to the procedure
   * ObjHookProc when the PC value equals ObjHook. The default value is 0.
   *
   * Note: the code for inserting the hook is commented out in exec.c
   * to speed-up execution. To use, uncomment it and rebuild aint.
   */

  ObjHook = 0;

  errflag = 0;
  while( (c = getopt(argc, argv, "x:hv")) != -1 ) {
    switch (c) {
      /*
	case 'C':
	Checkpoint_freq = strtol(optarg, NULL, 0);
	break;
	*/
    case 'j':
      job_separator = optarg;
      break;

    case 'x':
      { // to define scope for buf declaration
	char *buf = optarg;
	if ( (!strncmp(buf, "0x", 2)) || (!strncmp(buf, "0X", 2)) ) {
	  buf += 2;
	}
	
	istrstream istr(buf, sizeof(buf));
	istr >> ObjHook;
	informative( "ObjHookProc will be called at PC = " << hex << ObjHook << endl );
      }
      break;

    case 'h':
      usage(argc, argv);
      sim_usage(argc, argv);
      exit(0);

    case 'v':
      Verbose = 1;
      break;

      /* case 'w':
	 Print_Warnings = 1;
	 break;
	 
	 * case '-':
	 * break; */

    default:
      errflag = 1;
      break;
    }
  }

  // did we accidentally step on a "--"?
  /* FIXME: is this necessary?
  if ( (optind > 1) && (strncmp(argv[optind-1], "--", 2) == 0) ) {
    optind--;
  }
  */

  if (errflag) {
    usage(argc, argv);
    sim_usage(argc, argv);
    exit(1);
  }

}


