/* process.C
 * implementation of class process
 */

#include <iostream.h>
#include <string.h>
#include <String.h>
#include <sys/mman.h>
#include <sys/auxv.h>

#include "aint.h"
#include  "globals.h"
#include  "protos.h"

void load_object_file(process_ptr, const char *);
void object_file_info(const char *, aint_addr_t *, int *);

// Constructor used to initialize process from object file
process::process(const char *_Objname)
{
  Objname = strdup(_Objname);

  // Use the global N_procs to get the next higher pid
  pid = ProcessCount++;

  parent = youngest = sibling = NULL;

  tlb = new TLB(this);

  Shmem_regions = (struct Shm_ds *) NULL;
  Unsp_Shmat_Current = UNSP_SHMAT_START;

  segments = (segment_t *) NULL;
  n_segments = 0;

  n_instructions = 0;

  num_pages = num_private = 0;

  /* initialize all file descriptors to "closed" */
  for (int i=0; i<MAX_FDNUM; i++)  
    fd[i] = -1;

  pthread_keys_in_use = 0;

  sigv = new struct my_sigvec[33];

  // set stdin, stdout and stderr fds
  fd[0] = 0;
  fd[1] = 1;
  fd[2] = 2;

  // Initially, we are not related to any other process
  parent = youngest = sibling = (process_ptr) NULL;

  sim_user_addr = 0x0;

  object_file_info(Objname, &stack_end, &use_dynamic_loader);
  if (use_dynamic_loader) {
    const char *sbinloader = "/sbin/loader";
    informative("Executable " << Objname << ": using dynamic loader " << sbinloader << endl);
    load_object_file(this, sbinloader);
  } else {
    load_object_file(this, Objname);
  }

  // Create the main thread for this process
  thread_ptr pthread = new thread(this);
  threads = youngest_thread = pthread;
  thread_count = 1;

  // Create the address space for process/thread
  create_addr_space();

}


// constructor used to create child process from parent in fork()
process::process(process_ptr pprocess, thread_ptr pthread)
{
  pid = ProcessCount++;

  parent = pprocess;

  sibling = parent->youngest;

  parent->youngest = this;

  youngest = (process_ptr) NULL;

  // Create a thread corresponding to pthread in parent
  thread_ptr child_thread = new thread(this, pthread);

  threads = youngest_thread = child_thread;

  // Copy the memory state from parent
  // copy_addr_space(pprocess, pthread);

  // Copy file descriptors and the signal vector
  for (int i=0; i<MAX_FDNUM; i++) {
    // fd[j] contains the real (aint-space) filedescriptor that the object program can use
    fd[i] = pprocess->fd[i];
  }

  runstate = R_RUN;

}

// set up the address space structure of the process/thread
void
process::create_addr_space()
{
   /* the initial stack is below the text segment and grows down */ 
  /* be careful about programs compiled to use 32-bit addresses 
   * -- compute stack_end from process->entry_point
   */
  ulong mystack_end = BASE2ROUNDDOWN(stack_end, TB_PAGESIZE);
  ulong stack_size = Stack_size;
  int i;

  for (i = 0; i < n_segments; i++) {
    segment_t *segment = &segments[i];
    if (segment->start < mystack_end)
      mystack_end = segment->start;
  }

  if (Stack_size > mystack_end)
    stack_size = mystack_end/2;

  informative("\nEntryPoint: " << (void *) entry_point << endl);
  informative("\nSTACK_END: " << (void *) mystack_end << endl);

  /* magic number from Assembly Language Programmer's Guide */
  Data_end = 0x3ff00000000l;

  /* map the stack segment */
  add_segment(mystack_end - stack_size,
	      stack_size,
	      PROT_READ|PROT_WRITE,
	      MAP_PRIVATE);


  // Find the register SP is mapped to and set its value
  threads->path_tree->pMapReg(SP) = mystack_end;

  /* Set the gp register */
  threads->path_tree->pMapReg(GP) = initial_gp;
 
}


int
process::copy_argv(int arg_start, int argc, char **argv, char **envp)
{
  int next_arg;
  int argc_obj;
  const char *loader = "/sbin/loader";
  long auxv[8];
  thread_ptr pthread = threads;    // there is only one thread at this point
  char **eptr;
  int i;

  /* check for '--' separating jobs */

  // Determine argc for this job
  argc_obj = 0;

  for (next_arg=arg_start; next_arg<argc; next_arg++) {
    if ( (String) argv[next_arg] == (String) job_separator ) {
      next_arg++;
      break;
    }
    cout << argv[next_arg] << " ";
    argc_obj++;
  }
  cout << endl;

  /* add up sizes of all parameters */
  int size = 0;
  for (i=arg_start; i<argc_obj+arg_start; i++) {
    size += (strlen(argv[i]) + 1);
  }

  int nenv = 0;
  for (eptr = envp; *eptr; nenv++,eptr++) size += strlen(*eptr) + 1;

  /* leave space for the loader_filename */
  size += strlen(loader)+1;

  /* add in space for the argv and envp pointers, including NULLs */
  size += (argc_obj + nenv + 2) * sizeof(char *);
    
  size += sizeof(long); /* for argc */

  /* leave space for auxv */
  size += sizeof(auxv);

  /* Round up size to quad word boundary */
  size = BASE2ROUNDUP(size, 8);

  /* Allocate space on stack for the args */
  path_ptr ppath = pthread->path_tree;
  ppath->pMapReg(SP) -= size;

  /* sp = (long *) MAP(pthread->reg[SP]); */
  char *sp = (char *) ppath->pMapReg(SP);

  /* First item is argc */
  * (long *) PMAP((long) sp) = argc_obj; 
  sp += sizeof(long);

  /* Next comes the array of pointers argv */
  char **argv_obj = (char**)sp;
  /* Leave space for the argv array, including the NULL pointer */
  sp += (argc_obj + 1)*sizeof(char*);

  /* Pointer and space for the environment vars */
  char **envp_obj = (char**)sp;
  sp += (nenv + 1)*sizeof(char*);

  /* auxv */
  long *auxvp_obj = (long *)sp;
  sp += sizeof(auxv);

  /* where to write the strings */
  char *str_obj = (char *)sp;

  /* Copy the args to the stack of the main thread */
  informative("process " << pid << ": ");

  long exec_filename_obj = (long)str_obj;
  for (i = arg_start; i < (arg_start + argc_obj); i++) {
    informative(" " << argv[i]);

    strcpy_to_object_space((ulong)str_obj, argv[i]);

    * (char **) PMAP((long)(argv_obj + i - arg_start)) = str_obj;
    str_obj += strlen(argv[i]) + 1;
  }
  * (long **) PMAP((long)(argv_obj + argc_obj)) = NULL; // argv_obj is char**, argc_obj is int

  /* Copy the environment variables to the stack of the main thread */
  for (i = 0, eptr = envp; i < nenv; i++,eptr++) {

    strcpy_to_object_space((ulong)str_obj, (const char *) *eptr);

    * (char **) PMAP((long) (envp_obj + i)) = str_obj;
    str_obj += strlen(*eptr) + 1;
  }
  * (char **) PMAP((long) (envp_obj+ nenv)) = NULL;

  /* output the auxv */
  /* auxv filename */
  {
    long auxvi = 0;
    long exec_loader_filename_obj = (long)str_obj;
    auxv[auxvi++] = AT_EXEC_FILENAME;
    auxv[auxvi++] = exec_filename_obj;
    auxv[auxvi++] = AT_EXEC_LOADER_FILENAME;
    auxv[auxvi++] = exec_loader_filename_obj;
    auxv[auxvi++] = AT_EXEC_LOADER_FLAGS;
    auxv[auxvi++] = 0xbeef;
    auxv[auxvi++] = 0;
    auxv[auxvi++] = 0;

    strcpy_to_object_space(exec_loader_filename_obj, loader);
    memcpy_to_object_space((long)auxvp_obj, (char*)&auxv, sizeof(auxv));
  }

  return next_arg;

}


void
copy_addr_space(process_ptr pprocess, thread_ptr pthread)
{
}
