/* tlb.C
 * Implementation of aint's paged address space
 */


#include <iostream.h>
#include <assert.h>
#include <sys/mman.h>
#include <stdlib.h>

#include "aint.h"
#include "globals.h"
#include "protos.h"
#include "ipc.h"

// For an address not in process' tlb, allocate entry, and
// (if required) the page. Return tlb page table entry

tb_entry *
TLB::miss(aint_addr_t addr, long *paccess)
{
  tb_entry *pg;
  segment_t *segment;

  ulong vpage = TB_VPAGE(addr);
  ulong tbkey = TB_KEY(vpage);
  ulong tbtag = TB_TAG(vpage);

  
  /* TB MISS
   *   Check if address is private or shared.
   * If shared, look up global table for page.
   * If private, allocate here
   */


  /* Is it a valid private addr? */
  segment = pprocess->find_segment(addr);

  if (segment != NULL) {
    pg = new tb_entry;
    if (pg == NULL) {
      fatal("tlb_miss: cannot allocate private page table entry" << endl);
    }

    pg->vaddr = TB_VADDR(addr);
    pg->segment = segment;

    /* This goes in right at the beginning of the list */
    pg -> next = TB[tbkey];
    TB[tbkey] = pg;

    pg -> tag = tbtag;

    if (segment->flags & MAP_SHARED) {
      /* look for mmap() or shmat() attached memory */
      struct Shm_ds *shmds = pprocess->Shmem_regions;
      while (shmds) {
	if ((shmds->addr <= addr) && (addr < (shmds->addr + shmds->size))) {
	  ulong pageaddr = addr & PAGE_NUMBER_MASK;
	  pg -> page = Shmseg[shmds->shmid].paddr + (pageaddr - shmds->addr) ;

	  // The difference in bytes is divided by fe-vectors-per-page to get difference in page-fe-vectors
	  pg -> fe_vector = (unsigned long *) ( (void *)Shmseg[shmds->shmid].fe_vector +
	    ((pageaddr-shmds->addr) >> (TB_OFFSET_LENGTH-TB_FE_VEC_LENGTH)) );
	  break;
	}
	shmds = shmds->next;
      }
    }
    /* if it's not shared, or there's no attached segment, then allocate anonymous memory */
    if (! pg->page) {
      pg -> page = (void *) mmap( (caddr_t) NULL, TB_PAGESIZE, PROT_READ|PROT_WRITE,
				 MAP_ANONYMOUS|MAP_VARIABLE|MAP_PRIVATE, -1, 0);
      if (pg->page == (caddr_t) (-1)) fatal("tlb_miss: cannot allocate private page" << endl);

      /* Allocate the full-empty bit vector for this page. calloc initializes to empty (0) */
      pg->fe_vector = (unsigned long *) calloc(TB_FE_VEC_LONGS, sizeof(unsigned long));
      if (pg->fe_vector == NULL) fatal("tlb_miss: cannot allocate full-empty bit vector" << endl);
    }
    pg -> textpage = NULL;

    pg->flags = segment->flags;
    pg->prot = segment->prot;

    TB[tbkey]->lookaside = pg;
    /* Hack to speed-up copy_addr_space at fork-time */
    pprocess->num_pages++;
    pprocess->num_private++;

    if (paccess)
      *paccess = pg->prot;
    // return ((ulong) pg->page + TB_OFFSET(addr));
    return pg;
  }
  if (paccess)
    *paccess = 0;
  return (tb_entry *) NULL;    
}


/* addr2iphys(process, textaddr) returns the (struct icode *) 
 * pointer corresponding to textaddr.
 */
icode_ptr
process::addr2iphys(aint_addr_t addr, long *paccess)
{
  /* Refer to the Local TB First */
  tb_entry *pg;

  pg = tlb->Translate(addr, paccess);

  /* now check for corresponding textpage */
  if (pg) {
    if (!pg->textpage) {
      pg -> textpage = 
	(icode_ptr) mmap( (caddr_t) NULL, (TB_PAGESIZE >> 2) * sizeof(struct icode),
			 PROT_READ|PROT_WRITE,
			 MAP_ANONYMOUS|MAP_VARIABLE|MAP_PRIVATE, -1, 0);
#ifdef DEBUG
      fprintf(stderr, "addr2iphys(%lx) => [%p,%p]\n",
	      addr,
	      pg->textpage, pg->textpage + (TB_PAGESIZE >> 2) - 1);
#endif

      if (pg->textpage == (icode_ptr) (-1)) fatal("addr2iphys: cannot allocate private page" << endl);
    }
#ifdef DEBUG
    fprintf(stderr, "addr2iphys(%lx): pg=%p, textpage=%p, ITB_OFFSET=%lx, picode=%p\n",
	    addr, 
	    pg,
	    pg->textpage,
	    ITB_OFFSET(addr),
	    pg->textpage + ITB_OFFSET(addr));
#endif

    return (pg->textpage + ITB_OFFSET(addr));
  } else {
    return 0;
  }
}
